/*
 * Copyright (C) 2026 Samuel Weinig <sam@webkit.org>
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY APPLE INC. ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL APPLE INC. OR
 * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY
 * OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include "config.h"
#include "CSSKeywordValue.h"

#include <wtf/Hasher.h>

namespace WebCore {

CSSKeywordValue::CSSKeywordValue(StaticCSSValueTag, CSS::Keyword keyword)
    : CSSValue(ClassType::Keyword)
    , m_keyword(keyword)
{
    makeStatic();
}

CSSKeywordValue::CSSKeywordValue(StaticCSSValueTag, ImplicitInitialValueTag)
    : CSSKeywordValue(StaticCSSValue, CSS::Keyword { CSSValueInitial })
{
    m_isImplicitInitialValue = true;
}

String CSSKeywordValue::customCSSText(const CSS::SerializationContext& context) const
{
    return CSS::serializationForCSS(context, m_keyword);
}

bool CSSKeywordValue::equals(const CSSKeywordValue& other) const
{
    return m_keyword == other.m_keyword;
}

IterationStatus CSSKeywordValue::customVisitChildren(NOESCAPE const Function<IterationStatus(CSSValue&)>& func) const
{
    return CSS::visitCSSValueChildren(func, m_keyword);
}

bool CSSKeywordValue::addDerivedHash(Hasher& hasher) const
{
    add(hasher, m_keyword);
    return true;
}

String CSSKeywordValue::stringValue() const
{
    return nameStringForSerialization(m_keyword.value);
}

} // namespace WebCore
