/*
 * Copyright (C) 2026 Samuel Weinig <sam@webkit.org>
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY APPLE INC. ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL APPLE INC. OR
 * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY
 * OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#pragma once

#include "CSSPrimitiveNumericTypes.h"

namespace WebCore {
namespace CSS {

// <mask-border-width-value> = <length-percentage [0,∞]> | <number [0,∞]> | auto
struct MaskBorderWidthValue {
    using LengthPercentage = CSS::LengthPercentage<CSS::Nonnegative, float>;
    using Number = CSS::Number<CSS::Nonnegative, float>;

    MaskBorderWidthValue(Keyword::Auto keyword) : m_value { keyword } { }
    MaskBorderWidthValue(LengthPercentage&& lengthPercentage) : m_value { WTF::move(lengthPercentage) } { }
    MaskBorderWidthValue(Number&& number) : m_value { WTF::move(number) } { }

    bool isAuto() const { return WTF::holdsAlternative<Keyword::Auto>(m_value); }
    bool isLengthPercentage() const { return WTF::holdsAlternative<LengthPercentage>(m_value); }
    bool isNumber() const { return WTF::holdsAlternative<Number>(m_value); }

    template<typename... F> decltype(auto) switchOn(F&&... f) const
    {
        return WTF::switchOn(m_value, std::forward<F>(f)...);
    }

    bool operator==(const MaskBorderWidthValue&) const = default;

private:
    Variant<Keyword::Auto, LengthPercentage, Number> m_value;
};

// <'mask-border-width'> = [ <length-percentage [0,∞]> | <number [0,∞]> | auto ]{1,4}
// https://drafts.fxtf.org/css-masking-1/#propdef-mask-border-width
struct MaskBorderWidth {
    using Value = MaskBorderWidthValue;
    using Edges = MinimallySerializingSpaceSeparatedRectEdges<Value>;

    Edges values;

    bool NODELETE isInitialValue() const;

    bool operator==(const MaskBorderWidth&) const = default;
};
DEFINE_TYPE_WRAPPER_GET(MaskBorderWidth, values);

// MARK: - Conversion

template<> struct CSSValueCreation<MaskBorderWidth> { auto operator()(CSSValuePool&, const MaskBorderWidth&) -> Ref<CSSValue>; };

} // namespace CSS
} // namespace WebCore

DEFINE_TUPLE_LIKE_CONFORMANCE_FOR_TYPE_WRAPPER(WebCore::CSS::MaskBorderWidth)
DEFINE_VARIANT_LIKE_CONFORMANCE(WebCore::CSS::MaskBorderWidthValue)
