/**
 * Copyright (C) 2026 Igalia S.L.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 */

#pragma once

#include "RenderLayer.h"
#include "RenderSVGHiddenContainer.h"

namespace WebCore {

inline bool RenderLayer::isPaintingResourceLayerForSVG() const
{
    return m_svgData && m_svgData->isPaintingResourceLayer;
}

inline RenderSVGHiddenContainer* RenderLayer::enclosingHiddenOrResourceContainerForSVG() const
{
    return m_svgData ? m_svgData->enclosingHiddenOrResourceContainer.get() : nullptr;
}

} // namespace WebCore
